"""Loader class for downloading stock data.

Download stock data from yahoo finance.

"""


from collections import OrderedDict
from typing import Dict

import pandas as pd
import yfinance as yf

from mypo.market import Market


class Loader(object):
    """Loader class for downloading stock."""

    _tickers: Dict[str, pd.DataFrame]
    _expense_ratio: Dict[str, float]

    def __init__(self) -> None:  # pragma: no cover
        """Construct this object."""
        self._tickers = OrderedDict()
        self._expense_ratio = OrderedDict()

    def get(self, ticker: str, expense_ratio: float = 0.0) -> None:  # pragma: no cover
        """Get stock data of specified ticker.

        Args:
            ticker: Ticker that you want to download stock data.
            expense_ratio: Expense ratio of ticker. The default value is 0.0.

        Returns:
            Nothing
        """
        ticker = ticker.upper()
        df = yf.Ticker(ticker).history(period="max")
        df.index = pd.to_datetime(df.index)
        self._tickers[ticker] = df
        self._expense_ratio[ticker] = expense_ratio

    def get_market(self) -> Market:  # pragma: no cover
        """Get Market data.

        Returns:
            Market data
        """
        return Market.create_from_ticker(self._tickers, self._expense_ratio)



# flake8: noqa

from mypo.optimizer.base_optimizer import BaseOptimizer
from mypo.optimizer.cvar_optimizer import CVaROptimizer
from mypo.optimizer.minimum_variance_optimizer import MinimumVarianceOptimizer
from mypo.optimizer.no_optimizer import NoOptimizer
from mypo.optimizer.objective import covariance, semi_covariance
from mypo.optimizer.sharp_ratio_optimizer import SharpRatioOptimizer



"""Sampler class for generating scenarios."""
from __future__ import annotations

import pickle
from typing import List

import numpy as np
import pandas as pd
from numpy.random import multivariate_normal

from mypo.market import Market


class Sampler(object):
    """Sampler class for generating scenarios."""

    _mu: np.ndarray
    _chol: np.ndarray
    _columns: List[str]

    def __init__(self, market: Market, samples: int = 10) -> None:
        """Construct this object.

        Args:
            market: Market data.
            samples: Count of samples.
        """
        self.construct_model(market, samples)  # pragma: no cover

    def save(self, filepath: str) -> None:  # pragma: no cover
        """Save sampler data to file.

        Args:
            filepath: Path to file for storing data.
        """
        with open(filepath, "wb") as bin_file:
            pickle.dump(self, bin_file)

    @staticmethod
    def load(filepath: str) -> Sampler:
        """Load sampler data from file.

        Args:
            filepath: Path to file for loading data.

        Returns:
            Market object
        """
        with open(filepath, "rb") as bin_file:
            value: Sampler = pickle.load(bin_file)
            return value

    def construct_model(self, market: Market, samples: int = 10) -> None:  # pragma: no cover
        """Construct sampler model.

        Args:
            market: Market data.
            samples: Count of scenarios.
        """
        import warnings

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            import pymc3 as pm

        observed = market.get_rate_of_change()
        n = len(observed.columns)
        with pm.Model():
            # average
            prior_mu = pm.Uniform("prior_mu", -1, 1, shape=n)
            mu = pm.Normal("mu", mu=prior_mu, sigma=1, shape=n)

            sd_dist = pm.HalfCauchy.dist(beta=2.5)
            lower_triangular_chol = pm.LKJCholeskyCov("lower_triangular_chol", n=n, eta=1, sd_dist=sd_dist)

            chol = pm.expand_packed_triangular(n, lower_triangular_chol, lower=True)
            pm.MvNormal("observed_returns", mu=mu, chol=chol, observed=observed)

            trace = pm.sample(samples, pm.NUTS(), chains=3, return_inferencedata=False, random_seed=32)
            self._mu = trace["mu"]
            self._chol = trace["lower_triangular_chol"]
            self._columns = list(observed.columns)

    def sample(self, samples: int, seed: int = 32) -> pd.DataFrame:
        """Generate samples.

        Args:
            samples: Count of samples in a scenario.
            seed: Seed for random.

        Returns:
            Scenarios.
        """
        count_of_mu, _ = self._mu.shape
        np.random.seed(seed=seed)
        series = np.random.randint(0, count_of_mu, samples)
        prices = np.zeros(shape=(samples, len(self._columns)))
        for i, seed in enumerate(series):
            mu = self._mu[seed]
            cov = Sampler._get_covariance(self._chol[seed], len(self._columns))
            prices[i] = multivariate_normal(mu, cov, 1)[0]
        return pd.DataFrame(prices, columns=self._columns)

    @staticmethod
    def _get_covariance(elements: np.ndarray, n: int) -> np.ndarray:
        """Get symmetric matrix.

        Args:
            elements: Elements of low triangular matrix.
            n: Dimension of symmetric matrix.

        Returns:
            Symmetric matrix.
        """
        element = []
        for i in range(len(elements)):
            element += [elements[i]]
            element += [0 for _ in range(i + 1, n)]

        lower = np.array(element).reshape(2, 2)
        ret: np.ndarray = np.dot(lower, lower.T)
        return ret



"""Parameters for runner."""
from dataclasses import dataclass

import numpy as np


@dataclass
class Settings:
    """Settings for runner."""

    tax_rate: np.float64
    fee_rate: np.float64


DEFAULT_SETTINGS: Settings = Settings(tax_rate=np.float64(0.20), fee_rate=np.float64(0.005))



"""Loss functions."""

import numpy as np
import pandas as pd


def total_return(report: pd.DataFrame) -> np.float64:
    """Get negative total return.

    Args:
        report: Result of simulation.

    Returns:
        total return.
    """
    total_assets = report["total_assets"]
    return np.float64(total_assets[len(total_assets) - 1] / total_assets[0])


def yearly_total_return(report: pd.DataFrame, frequency: int = 252) -> np.float64:
    """Get negative total return.

    Args:
        report: Result of simulation.
        frequency: The count of days of trading.

    Returns:
        yearly total return.
    """
    return total_return(report) ** (frequency / len(report))


def max_drawdown(report: pd.DataFrame) -> np.float64:
    """Get negative total return.

    Args:
        report: Result of simulation.

    Returns:
        Negative total return.
    """
    max_assets = report["total_assets"].cummax()
    return np.float64(np.min(report["total_assets"] / max_assets))


def max_drawdown_span(report: pd.DataFrame) -> int:
    """Get negative total return.

    Args:
        report: Result of simulation.

    Returns:
        Negative total return.
    """
    df = report[["total_assets"]].copy()
    df["max_total"] = df["total_assets"] < df["total_assets"].cummax()
    df["continuous"] = df.groupby((df["max_total"] != df["max_total"].shift()).cumsum()).cumcount() + 1
    df = df[df["max_total"]]
    ret: int = np.max(df["continuous"])
    return ret



"""Utility functions for model selection."""
from typing import List, Tuple

from mypo.market import Market


def split_n_periods(market: Market, n: int, train_span: int) -> Tuple[List[Market], List[Market]]:
    """Split market to n periods.

    Args:
        market: Market data
        n: Count of split
        train_span: Train span if you want to specify.

    Returns:
        Split market data
    """
    index = market.get_index()
    len_index = len(index)

    eval_span = int((len_index - train_span) / n)
    train_market: List[Market] = []
    eval_market: List[Market] = []
    for i in range(n):
        start_eval_index = i * eval_span + train_span
        train_market += [market.extract(index[start_eval_index - train_span : start_eval_index])]
        eval_market += [market.extract(index[start_eval_index : start_eval_index + eval_span])]

    return train_market, eval_market



"""Rebalance strategies."""
from datetime import datetime

import numpy as np
import numpy.typing as npt

from mypo import Market
from mypo.common import safe_cast
from mypo.optimizer import BaseOptimizer, NoOptimizer
from mypo.trigger import BaseTrigger


class BaseRebalancer(object):
    """Interface class of Rebalance strategy class."""

    _trigger: BaseTrigger
    _optimizer: BaseOptimizer
    _do_re_optimize: bool

    def __init__(
        self, trigger: BaseTrigger, optimizer: BaseOptimizer = NoOptimizer(), do_re_optimize: bool = False
    ) -> None:
        """Construct object.

        Args:
            trigger: Trigger.
            optimizer: Optimizer.
            do_re_optimize: Re-optimize if it's True. The default is False.
        """
        self._trigger = trigger
        self._optimizer = optimizer
        self._do_re_optimize = do_re_optimize

    def apply(self, at: datetime, market: Market, assets: npt.ArrayLike, cash: np.float64) -> np.ndarray:
        """Apply rebalance strategy to current situation.

        Args:
            at: Current date for applying rebalance.
            market: Market data for all spans.
            assets: Current assets for applying rebalance.
            cash: Current cash for applying rebalance.

        Returns:
            Deal
        """
        assets = safe_cast(assets)
        if self._trigger.is_fire(at, assets, cash, self._optimizer.get_weights()):
            if self._do_re_optimize:
                self._optimizer.optimize(market, at)
            diff = self._rebalance(assets)
        else:
            diff = self._do_nothing(assets)
        if cash < 0 and np.sum(assets + diff) > 0:
            diff = diff - np.abs(cash) * self._optimizer.get_weights()
        return diff

    def _rebalance(self, assets: np.ndarray) -> np.ndarray:
        diff: np.ndarray = self._optimizer.get_weights() * np.sum(assets) - assets
        return diff

    def _do_nothing(self, assets: np.ndarray) -> np.ndarray:
        zero: np.ndarray = assets - assets
        return zero



"""Rebalance strategies."""

import numpy as np

from mypo.optimizer import BaseOptimizer
from mypo.rebalancer import BaseRebalancer
from mypo.trigger import ThresholdTrigger


class ThresholdRebalancer(BaseRebalancer):
    """Weighted rebalance strategy by monthly applying."""

    _threshold: np.float64

    def __init__(
        self, optimizer: BaseOptimizer, do_re_optimize: bool = False, threshold: np.float64 = np.float64(0.05)
    ) -> None:
        """Construct object.

        Args:
            optimizer: Optimizer.
            do_re_optimize: Re-optimize if it's True. The default is False.
            threshold: Threshold of fire.
        """
        super().__init__(
            trigger=ThresholdTrigger(threshold=threshold), optimizer=optimizer, do_re_optimize=do_re_optimize
        )



"""Rebalance strategies."""
import datetime

import numpy as np
import numpy.typing as npt

from mypo.trigger.base_trigger import BaseTrigger


class MonthlyTrigger(BaseTrigger):
    """Weighted rebalance strategy by monthly applying."""

    _old_month: int
    _weights: np.ndarray

    def __init__(self, old_month: int = 0) -> None:
        """Construct object.

        Args:
        old_month: Previous month.
        """
        super().__init__()
        self._old_month = old_month

    def is_fire(
        self,
        index: datetime.datetime,
        assets: npt.ArrayLike,
        cash: np.float64,
        weights: npt.ArrayLike,
    ) -> bool:
        """Apply rebalance strategy to current situation.

        Args:
            index: Current date for applying rebalance.
            assets: Current assets for applying rebalance.
            cash: Current cash for applying rebalance.
            weights: Weights of assets.

        Returns:
            Deal
        """
        if self._old_month != index.month:
            self._old_month = index.month
            return True
        else:
            return False



"""Rebalance strategies."""

import numpy.typing as npt

from mypo.optimizer import NoOptimizer
from mypo.rebalancer import BaseRebalancer
from mypo.trigger import AlwaysTrigger


class PlainRebalancer(BaseRebalancer):
    """Simple weighted rebalance strategy."""

    def __init__(self, weights: npt.ArrayLike) -> None:
        """Construct object.

        Args:
            weights: Weights.
        """
        super().__init__(trigger=AlwaysTrigger(), optimizer=NoOptimizer(weights))



"""Rebalance strategies."""
import datetime

import numpy as np
import numpy.typing as npt

from mypo.common import safe_cast
from mypo.trigger.base_trigger import BaseTrigger


class ThresholdTrigger(BaseTrigger):
    """Weighted rebalance strategy by monthly applying."""

    _threshold: np.float64
    _weights: np.ndarray

    def __init__(self, threshold: np.float64 = np.float64(0.05)) -> None:
        """Construct object.

        Args:
            threshold: Threshold of fire.
        """
        self._threshold = threshold

    def is_fire(
        self,
        index: datetime.datetime,
        assets: npt.ArrayLike,
        cash: np.float64,
        weights: npt.ArrayLike,
    ) -> bool:
        """Apply rebalance strategy to current situation.

        Args:
            index: Current date for applying rebalance.
            assets: Current assets for applying rebalance.
            cash: Current cash for applying rebalance.
            weights: Weights of assets.

        Returns:
            Deal
        """
        assets = safe_cast(assets)
        if np.max(np.abs(assets / np.sum(assets) - weights)) > self._threshold:
            return True
        else:
            return False



"""Optimizer for weights of portfolio."""
from datetime import datetime
from typing import Optional

import numpy.typing as npt

from mypo import Market
from mypo.optimizer import BaseOptimizer


class NoOptimizer(BaseOptimizer):
    """Base Optimizer."""

    def __init__(self, weights: Optional[npt.ArrayLike] = None):
        """Construct this object.

        Args:
            weights: Weight.
        """
        super().__init__(weights)

    def optimize(self, market: Market, at: datetime) -> None:
        """Optimize weights.

        Args:
            market: Market data.
            at: Current date.
        """
        pass



# flake8: noqa

from mypo.trigger.always_trigger import AlwaysTrigger
from mypo.trigger.base_trigger import BaseTrigger
from mypo.trigger.monthly_trigger import MonthlyTrigger
from mypo.trigger.no_trigger import NoTrigger
from mypo.trigger.threshold_trigger import ThresholdTrigger



"""Optimizer for weights of portfolio."""

from datetime import datetime
from typing import Optional

import numpy as np
from scipy.optimize import minimize

from mypo.common import safe_cast
from mypo.market import Market
from mypo.optimizer.base_optimizer import BaseOptimizer
from mypo.optimizer.objective import covariance, semi_covariance


class MinimumVarianceOptimizer(BaseOptimizer):
    """Minimum variance optimizer."""

    _with_semi_covariance: bool
    _span: int

    def __init__(
        self,
        span: int = 260,
        with_semi_covariance: bool = False,
        minimum_return: Optional[float] = None,
    ):
        """Construct this object.

        Args:
            span: Span for evaluation.
            with_semi_covariance: whether use semi covariance mode if it's Ture.
            minimum_return: Minimum return.
        """
        self._span = span
        self._with_semi_covariance = with_semi_covariance
        self._minimum_return = minimum_return
        super().__init__([1])

    def optimize(self, market: Market, at: datetime) -> None:
        """Optimize weights.

        Args:
            market: Past market stock prices.
            at: Current date.

        Returns:
            Optimized weights
        """
        historical_data = market.extract(market.get_index() < at).get_rate_of_change()
        prices = historical_data.tail(n=self._span).to_numpy()
        Q = semi_covariance(prices) if self._with_semi_covariance else covariance(prices)
        n = len(historical_data.columns)
        x = np.ones(n) / n

        def fn(x: np.ndarray, Q: np.ndarray) -> np.float64:
            ret: np.float64 = np.dot(np.dot(x, Q), x.T) / np.max(np.abs(Q))
            return ret

        cons = [{"type": "eq", "fun": lambda x: np.sum(x) - 1}]
        if self._minimum_return is not None:
            ret = prices.mean(axis=0)
            daily_risk_free_rate = (1.0 + self._minimum_return) ** (1 / 252) - 1.0
            cons += [
                {
                    "type": "ineq",
                    "fun": lambda x: np.dot(ret, x) - daily_risk_free_rate,
                },
            ]

        bounds = [[0.0, 1.0] for i in range(n)]
        minout = minimize(fn, x, args=(Q), method="SLSQP", bounds=bounds, constraints=cons)
        self._weights = safe_cast(minout.x)



"""Optimizer for weights of portfolio."""
from datetime import datetime
from typing import Optional

import numpy as np
from scipy.optimize import minimize

from mypo.common import safe_cast
from mypo.market import Market
from mypo.optimizer.base_optimizer import BaseOptimizer
from mypo.sampler import Sampler


class CVaROptimizer(BaseOptimizer):
    """Minimum variance optimizer."""

    _span: int
    _beta: float
    _samples: int
    _sampler: Optional[Sampler]

    def __init__(self, span: int = 260, beta: float = 0.05, samples: int = 10, sampler: Optional[Sampler] = None):
        """Construct this object.

        Args:
            sampler: Sampler.
            span: Span for evaluation.
            beta: Confidence.
            samples: Count of scenarios.
            sampler: Sampler.
        """
        self._span = span
        self._beta = beta
        self._samples = samples
        self._sampler = sampler
        super().__init__([1])

    def optimize(self, market: Market, at: datetime) -> None:
        """Optimize weights.

        Args:
            market: Past market stock prices.
            at: Current date.

        Returns:
            Optimized weights
        """
        historical_data = market.extract(market.get_index() < at).tail(self._span)
        sampler = Sampler(market=historical_data, samples=self._samples) if self._sampler is None else self._sampler
        sample = sampler.sample(self._span).to_numpy()

        n = len(historical_data.get_tickers())
        x = np.ones(n) / n

        def fn(x: np.ndarray, sequence: np.ndarray) -> np.float64:
            r = np.dot(sequence, x.T) / np.max(sequence)
            return -np.float64(np.mean(np.where(r < np.quantile(r, self._beta), r, 0)))

        cons = {"type": "eq", "fun": lambda x: np.sum(x) - 1}
        bounds = [[0.0, 1.0] for i in range(n)]
        minout = minimize(fn, x, args=(sample), method="SLSQP", bounds=bounds, constraints=cons)
        self._weights = safe_cast(minout.x)



"""Simulation."""

from typing import Any, List, Optional

import numpy as np
import pandas as pd
from tqdm import tqdm

from mypo.common import calc_capital_gain_tax, calc_fee, calc_income_gain_tax, safe_cast
from mypo.market import Market
from mypo.rebalancer.base_rebalancer import BaseRebalancer
from mypo.rebalancer.no_rebalancer import NoRebalancer
from mypo.reporter import Reporter
from mypo.settings import DEFAULT_SETTINGS, Settings

WEEK_DAYS = int(365 * 5 / 7)


class Runner(object):
    """Runner of simulation."""

    _assets: np.ndarray
    _average_assets_prices: np.ndarray
    _rebalancer: BaseRebalancer
    _reporter: Reporter
    _cash: np.float64
    _withdraw: np.float64
    _settings: Settings
    _month: int

    def __init__(
        self,
        assets: Optional[List[float]] = None,
        cash: float = 0.0,
        withdraw: float = 0.0,
        rebalancer: BaseRebalancer = NoRebalancer(),
        settings: Settings = DEFAULT_SETTINGS,
    ):
        """Construct this object.

        Args:
            assets: Initial asset.
            rebalancer: Rebalance strategy.
            cash: Initial cash.
            withdraw: Withdraw.
            settings: Settings.
        """
        if assets is None:
            assets = [1e-23]
        self._assets = safe_cast(assets)
        self._average_assets_prices = np.ones(len(self._assets))
        self._rebalancer = rebalancer
        self._reporter = Reporter()
        self._cash = np.float64(cash)
        self._withdraw = np.float64(withdraw)
        self._settings = settings
        self._month = 0

    def total_assets(self) -> np.float64:
        """Get current total assets. Total asset is addition of stock assets and cash.

        Returns:
            Total assets.
        """
        return np.float64(np.sum(self._assets) + self._cash)

    def apply(self, market: Market, i: int) -> None:
        """Apply current market situation.

        Args:
            market: Market data.
            i: Index.
        """
        at = market.get_index()[i]
        prices = market.get_rate_of_change().to_records(index=False)[i]
        price_dividends_yield = market.get_price_dividends_yield().to_records(index=False)[i]
        expense_ratio = market.get_expense_ratio()

        prices = safe_cast(prices)
        price_dividends_yield = safe_cast(price_dividends_yield)
        expense_ratio = safe_cast(expense_ratio)

        # apply withdraw
        if self._month != at.month:
            self._cash -= self._withdraw / 12
        self._month = at.month

        # apply market prices
        previous_assets = self._assets
        self._assets = self._assets * (1.0 + prices)
        capital_gain = np.float64(self._assets.sum() - previous_assets.sum())

        diff = self._rebalancer.apply(at, market, self._assets, self._cash)
        self._assets += diff
        capital_gain_tax = calc_capital_gain_tax(self._average_assets_prices, prices, diff, self._settings)
        fee = calc_fee(diff, self._settings)

        # process of income gain
        income_gain = (self._assets * price_dividends_yield).sum()
        income_gain_tax = calc_income_gain_tax(self._assets, price_dividends_yield, self._settings)

        # process of others
        self._cash = self._cash + income_gain - income_gain_tax - diff.sum() - fee - capital_gain_tax
        self._assets = (1.0 - expense_ratio / WEEK_DAYS) * self._assets
        deal = np.max([diff.sum(where=diff > 0), -diff.sum(where=diff < 0)])
        trading_prices = np.where(diff > 0, 1.0 + prices, self._average_assets_prices)
        self._average_assets_prices = (
            self._average_assets_prices * previous_assets + diff * trading_prices
        ) / self._assets

        # record to reporter
        self._reporter.record(
            at,
            self.total_assets(),
            capital_gain,
            income_gain,
            self._cash,
            deal,
            fee,
            capital_gain_tax,
            income_gain_tax,
        )

    def run(self, market: Market, train_span: int = 0, verbose: bool = False) -> None:
        """Run simulation.

        Args:
            market: Market data.
            train_span: Periods of training span.
            verbose: Show progress.
        """
        target = range(train_span, market.get_length())

        def wrap(x: Any) -> Any:
            """Wrapper for tqdm."""
            return tqdm(x) if verbose else x

        for i in wrap(target):
            self.apply(market, i)

    def report(self) -> pd.DataFrame:
        """Report simulation.

        Returns:
            result
        """
        return self._reporter.report()



"""Market object for store and loading stock prices data."""

from __future__ import annotations

import pickle
from datetime import datetime
from enum import Enum
from typing import Dict, List

import numpy as np
import pandas as pd

from mypo.common import safe_cast


class SamplingMethod(Enum):
    """Enum of method for resampling."""

    YEAR = "Y"
    MONTH = "M"


class Market(object):
    """Market class for store and loading stock prices data."""

    _expense_ratio: Dict[str, float]
    _closes: pd.DataFrame
    _price_dividends_yield: pd.DataFrame

    def __init__(self, closes: pd.DataFrame, price_dividends_yield: pd.DataFrame, expense_ratio: Dict[str, float]):
        """Construct this object.

        Args:
            closes: Tickers.
            price_dividends_yield: Price dividends yield.
            expense_ratio: Expense ratio.
        """
        self._expense_ratio = expense_ratio
        self._closes = closes
        self._price_dividends_yield = price_dividends_yield

    def save(self, filepath: str) -> None:  # pragma: no cover
        """Save market data to file.

        Args:
            filepath: Path to file for storing data.
        """
        with open(filepath, "wb") as bin_file:
            pickle.dump(self, bin_file)

    @staticmethod
    def load(filepath: str) -> Market:
        """Load market data from file.

        Args:
            filepath: Path to file for loading data.

        Returns:
            Market object
        """
        with open(filepath, "rb") as bin_file:
            value: Market = pickle.load(bin_file)
            return value

    @staticmethod
    def create(start_date: str, end_date: str, yearly_gain: float, ticker: str = "None") -> Market:
        """Load market data from file.

        Args:
            ticker: Ticker.
            start_date: Start date.
            end_date: End date.
            yearly_gain: Yearly gain.

        Returns:
            Market object
        """
        index = pd.date_range(start_date, end_date, freq="D")
        n = len(index)
        daily_gain = (1.0 + yearly_gain) ** (1 / 365)
        prices = np.ones(n) * (daily_gain ** np.arange(n))
        return Market.create_from_ticker(
            tickers={ticker: pd.DataFrame({"Close": prices, "Dividends": np.zeros(n)}, index=index)},
            expense_ratio={ticker: 0.0},
        )

    @staticmethod
    def create_from_ticker(tickers: Dict[str, pd.DataFrame], expense_ratio: Dict[str, float]) -> Market:
        """Construct this object.

        Args:
            tickers: Tickers.
            expense_ratio: Expense ratio.
        """
        return Market(
            closes=Market.calc_raw(tickers),
            price_dividends_yield=Market.calc_price_dividends_yield(tickers),
            expense_ratio=expense_ratio,
        )

    @staticmethod
    def calc_raw(tickers: Dict[str, pd.DataFrame]) -> pd.DataFrame:
        """Get price data from stored market data.

        Args:
            tickers: Tickers

        Returns:
            Prices
        """
        rs = [tickers[ticker][["Close"]] for ticker in tickers.keys()]
        df = pd.concat(rs, axis=1, join="inner")
        df.columns = tickers.keys()
        return df

    @staticmethod
    def calc_price_dividends_yield(tickers: Dict[str, pd.DataFrame]) -> pd.DataFrame:
        """Get price dividends yield.

        Returns:
            price dividends yield data
        """
        rs = [tickers[ticker]["Dividends"] / tickers[ticker]["Close"] for ticker in tickers.keys()]
        df = pd.concat(rs, axis=1, join="inner")
        df.columns = tickers.keys()
        return df

    def resample(self, method: SamplingMethod) -> Market:
        """Resample market data.

        Args:
            method: Frequency of sampling.

        Returns:
            Resampled data.
        """
        market = self.trim(method)
        if method == SamplingMethod.YEAR:
            rule = "Y"
        elif method == SamplingMethod.MONTH:
            rule = "M"
        else:
            assert False  # pragma: no cover
        return Market(
            closes=market._closes.resample(str(rule)).last(),
            price_dividends_yield=market._price_dividends_yield.resample(rule).sum(),
            expense_ratio=market._expense_ratio,
        )

    def trim(self, method: SamplingMethod) -> Market:
        """Trim market data.

        Args:
            method: Frequency of sampling.

        Returns:
            Trim data.
        """
        if method == SamplingMethod.YEAR:
            rule = "Y"
        elif method == SamplingMethod.MONTH:
            rule = "M"
        else:
            assert False  # pragma: no cover
        df = self._closes.groupby(pd.Grouper(freq=rule)).sum()
        if method == SamplingMethod.YEAR:
            first = df.index[0] if self._closes.index.is_year_start[0] else df.index[1]
            last = df.index[-1] if self._closes.index.is_year_end[-1] else df.index[-2]
        elif method == SamplingMethod.MONTH:
            first = df.index[0] if self._closes.index.is_month_start[0] else df.index[1]
            last = df.index[-1] if self._closes.index.is_month_end[-1] else df.index[-2]
        else:
            assert False  # pragma: no cover
        return Market(
            closes=self._closes[first:last],
            price_dividends_yield=self._price_dividends_yield[first:last],
            expense_ratio=self._expense_ratio,
        )

    def extract(self, index: pd.Series) -> Market:
        """Extract market data.

        Args:
            index: Index what you want to extract.

        Returns:
            Extracted Data
        """
        return Market(
            closes=self._closes.loc[index],
            price_dividends_yield=self._price_dividends_yield.loc[index],
            expense_ratio=self._expense_ratio,
        )

    def tail(self, n: int) -> Market:
        """Extract market data.

        Args:
            n: Last n records.

        Returns:
            Extracted Data
        """
        return Market(
            closes=self._closes.loc[self._closes.tail(n).index],
            price_dividends_yield=self._price_dividends_yield.loc[self._price_dividends_yield.tail(n).index],
            expense_ratio=self._expense_ratio,
        )

    def head(self, n: int) -> Market:
        """Extract market data.

        Args:
            n: First n records.

        Returns:
            Extracted Data
        """
        return Market(
            closes=self._closes.loc[self._closes.head(n).index],
            price_dividends_yield=self._price_dividends_yield.loc[self._price_dividends_yield.head(n).index],
            expense_ratio=self._expense_ratio,
        )

    def get_raw(self) -> pd.DataFrame:
        """Get price data from stored market data.

        Returns:
            Prices
        """
        return self._closes

    def get_tickers(self) -> List[str]:
        """Get tickers.

        Returns:
            Tickers.
        """
        return list(self._closes.columns)

    def get_length(self) -> int:
        """Get length.

        Returns:
            length.
        """
        return len(self.get_rate_of_change())

    def get_index(self) -> pd.Index:
        """Get index date from stored market data.

        Returns:
            index date
        """
        return self._closes.index

    def get_first_date(self) -> datetime:
        """Get first date.

        Returns:
            First date.
        """
        first_date: datetime = self.get_index()[0]
        return first_date

    def get_last_date(self) -> datetime:
        """Get Last date.

        Returns:
            Last date.
        """
        last_date: datetime = self.get_index()[-1]
        return last_date

    def get_normalized_prices(self) -> pd.DataFrame:
        """Get normalized prices.

        Returns:
            Normalized prices
        """
        df = self._closes
        for c in df.columns:
            df[c] = df[c] / df[c][0]
        return df

    def get_rate_of_change(self) -> pd.DataFrame:
        """Get rate of change of prices.

        Returns:
            Rate of change
        """
        df = self._closes
        df = df.pct_change(axis=0)
        df.dropna(inplace=True)
        return df

    def get_price_dividends_yield(self) -> pd.DataFrame:
        """Get price dividends yield.

        Returns:
            price dividends yield data
        """
        return self._price_dividends_yield

    def get_expense_ratio(self) -> np.ndarray:
        """Get expense ratio.

        Returns:
            Expense ratio
        """
        rs = [self._expense_ratio[ticker] for ticker in self._closes.columns]
        return safe_cast(rs)



"""Utility functions.

A collection of utlities by using in Runner class.

"""


from typing import Any

import numpy as np
import numpy.typing as npt

from mypo.settings import Settings


def safe_cast(value: Any) -> np.ndarray:
    """Cast array like type to numpy array.

    Args:
        value: target value

    Returns:
        casted_value
    """
    return np.array(list(value), dtype=np.float64)


def calc_capital_gain_tax(
    average_asset_prices: npt.ArrayLike, prices: npt.ArrayLike, diff: npt.ArrayLike, settings: Settings
) -> np.float64:
    """Calculate capital gain from market data and assets.

    Args:
        average_asset_prices: average assets from investment started.
        prices: Current assets.
        diff: Deal.
        settings: Ratio of tax for capital gain.

    Returns:
        capital_gain_tax
    """
    average_asset_prices = safe_cast(average_asset_prices)
    prices = safe_cast(prices)
    diff = safe_cast(diff)
    deal_sell = np.where(diff > 0, 0, diff)
    unrealized_gain = prices - average_asset_prices
    plus_unrealized_gain = np.where(unrealized_gain > 0, unrealized_gain, 0)
    capital_gain_tax: np.float64 = np.sum(plus_unrealized_gain * deal_sell * settings.tax_rate)
    return capital_gain_tax


def calc_income_gain_tax(assets: npt.ArrayLike, price_dividends_yield: npt.ArrayLike, settings: Settings) -> np.float64:
    """Calculate income gain from market data and assets.

    Args:
        assets: Current assets.
        price_dividends_yield: Price dividends yield.
        settings: Ratio of tax for capital gain.

    Returns:
        income_gain_tax
    """
    assets = safe_cast(assets)
    price_dividends_yield = safe_cast(price_dividends_yield)
    income_gain_tax: np.float64 = -np.sum(assets * price_dividends_yield * settings.tax_rate)
    return income_gain_tax


def calc_fee(diff: npt.ArrayLike, settings: Settings) -> np.float64:
    """Calculate fee for dealing.

    Args:
        diff: Deal.
        settings: Ratio of fee for trading.

    Returns:
        fee
    """
    diff = safe_cast(diff)
    deal = np.abs(diff)
    fee: np.float64 = -np.sum(deal * settings.fee_rate)
    return fee



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mypo',
 'mypo.indicator',
 'mypo.optimizer',
 'mypo.rebalancer',
 'mypo.sampler',
 'mypo.trigger']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19.2,<2.0.0',
 'pandas>=1.1.0,<2.0.0',
 'pymc3>=3.11.1,<4.0.0',
 'scipy>=1.6.0,<2.0.0',
 'tqdm>=4.59.0,<5.0.0',
 'yfinance>=0.1.55,<0.2.0']

setup_kwargs = {
    'name': 'mypo',
    'version': '0.0.69',
    'description': '',
    'long_description': "mypo is a Python module for making strategy your portfolio and is distributed under the MIT license.\n\nInstallation\n============\n\nYou can install the latest this module with the command:\n\n\n    pip install mypo\n\nQuick start\n============\n\nYou can start quickly  by using docker image with the command.\n\n\n    git clone https://github.com/sonesuke/mypo.git\n    docker-compose up\n\nAfter starting Jupyter server, you can login it with the token. The default of token is '1111'.\n\n\nImportant links\n===============\n\n- Documentation: https://sonesuke.github.io/mypo\n- Official source code repo: https://github.com/sonesuke/mypo\n- Download releases: https://pypi.org/project/mypo/\n- Issue tracker: https://github.com/sonesuke/mypo/issues\n\nCitation\n========\nIf you use mypo in a publication, we would appreciate citations: https://sonesuke.github.io/mypo/",
    'author': 'sonesuke',
    'author_email': 'iamsonesuke@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sonesuke/mypo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.10,<4.0.0',
}


setup(**setup_kwargs)



"""Rebalance strategies."""
import datetime

import numpy as np
import numpy.typing as npt


class BaseTrigger(object):
    """Weighted rebalance strategy by monthly applying."""

    def is_fire(
        self,
        index: datetime.datetime,
        assets: npt.ArrayLike,
        cash: np.float64,
        weights: npt.ArrayLike,
    ) -> bool:
        """Apply rebalance strategy to current situation.

        Args:
            index: Current date for applying rebalance.
            assets: Current assets for applying rebalance.
            cash: Current cash for applying rebalance.
            weights: Weights of assets.

        Returns:
            Deal
        """
        pass  # pragma: no cover



"""Optimizer for weights of portfolio."""

from datetime import datetime

import numpy as np
from scipy.optimize import minimize

from mypo.common import safe_cast
from mypo.market import Market
from mypo.optimizer import BaseOptimizer
from mypo.optimizer.objective import sharp_ratio


class SharpRatioOptimizer(BaseOptimizer):
    """Minimum variance optimizer."""

    _span: int
    _risk_free_rate: float

    def __init__(
        self,
        risk_free_rate: float = 0.02,
        span: int = 260,
    ):
        """Construct this object.

        Args:
            risk_free_rate: Risk free rate
            span: Span for evaluation.
        """
        self._risk_free_rate = risk_free_rate
        self._span = span
        super().__init__([1])

    def optimize(self, market: Market, at: datetime) -> None:
        """Optimize weights.

        Args:
            market: Past market stock prices.
            at: Current date.

        Returns:
            Optimized weights
        """
        historical_data = market.extract(market.get_index() < at).get_rate_of_change()
        prices = historical_data.tail(n=self._span).to_numpy()
        Q = np.cov(prices.T)
        R = prices.mean(axis=0).T
        n = Q.shape[0]
        x = np.ones(n) / n
        daily_risk_free_rate = (1.0 + self._risk_free_rate) ** (1 / 252) - 1.0

        def fn(
            x: np.ndarray,
            R: np.ndarray,
            Q: np.ndarray,
            daily_risk_free_rate: np.float64,
        ) -> np.float64:
            return -sharp_ratio(np.dot(x, R), np.dot(np.dot(x, Q), x.T), daily_risk_free_rate)

        cons = {"type": "eq", "fun": lambda x: np.sum(x) - 1}
        bounds = [[0.0, 1.0] for i in range(n)]
        minout = minimize(
            fn,
            x,
            args=(R, Q, daily_risk_free_rate),
            method="SLSQP",
            bounds=bounds,
            constraints=cons,
        )
        self._weights = safe_cast(minout.x)



"""Rebalance strategies."""

from mypo.optimizer import BaseOptimizer
from mypo.rebalancer import BaseRebalancer
from mypo.trigger import MonthlyTrigger


class MonthlyRebalancer(BaseRebalancer):
    """Weighted rebalance strategy by monthly applying."""

    def __init__(self, optimizer: BaseOptimizer, do_re_optimize: bool = False, old_month: int = 0) -> None:
        """Construct object.

        Args:
            optimizer: Optimizer.
            do_re_optimize: Re-optimize if it's True. The default is False.
            old_month: Previous month.
        """
        super().__init__(trigger=MonthlyTrigger(old_month), optimizer=optimizer, do_re_optimize=do_re_optimize)



# flake8: noqa

from mypo.rebalancer.base_rebalancer import BaseRebalancer
from mypo.rebalancer.monthly_rebalancer import MonthlyRebalancer
from mypo.rebalancer.no_rebalancer import NoRebalancer
from mypo.rebalancer.plain_rebalancer import PlainRebalancer
from mypo.rebalancer.threshold_rebalancer import ThresholdRebalancer



"""Optimizer for weights of portfolio."""
from datetime import datetime
from typing import Optional

import numpy as np
import numpy.typing as npt

from mypo import Market
from mypo.common import safe_cast


class BaseOptimizer(object):
    """Base Optimizer."""

    _weights: np.ndarray

    def __init__(self, weights: Optional[npt.ArrayLike]):
        """Construct this object.

        Args:
            weights: Weight.
        """
        if weights is None:
            weights = [1]
        self._weights = safe_cast(weights)

    def get_weights(self) -> np.ndarray:
        """Get weights.

        Returns:
            Weights.
        """
        return self._weights

    def optimize(self, market: Market, at: datetime) -> None:
        """Optimize weights.

        Args:
            market: Market data.
            at: Current date.
        """
        pass  # pragma: no cover



"""Covariance and semi-covariance."""

import numpy as np


def covariance(prices: np.ndarray) -> np.ndarray:
    """Calculate covariance matrix.

    Args:
        prices: Prices of market data.

    Returns:
        Covariance matrix.
    """
    Q = np.cov(prices.T)
    return np.array(Q)


def semi_covariance(prices: np.ndarray) -> np.ndarray:
    """Calculate semi-covariance matrix.

    Args:
        prices: Prices of market data.

    Returns:
        Semi-covariance matrix.
    """
    Q = np.cov(np.where(prices.T < 0, prices.T, 0.0))
    return np.array(Q)


def sharp_ratio(r: np.float64, q: np.float64, risk_free_rate: np.float64) -> np.float64:
    """Calculate Sharp ratio.

    Args:
        r: Return
        q: Variance
        risk_free_rate: Risk free rate

    Returns:
        Sharp ratio.
    """
    return (r - risk_free_rate) / q



"""Rebalance strategies."""

from mypo.optimizer import NoOptimizer
from mypo.rebalancer import BaseRebalancer
from mypo.trigger import NoTrigger


class NoRebalancer(BaseRebalancer):
    """Simple weighted rebalance strategy."""

    def __init__(self) -> None:
        """Construct object."""
        super().__init__(trigger=NoTrigger(), optimizer=NoOptimizer())
